import logging
import ssl
from collections import defaultdict
from enum import Enum, auto
from typing import Awaitable

import certifi
import websockets
from websockets.exceptions import ConnectionClosedError

from hasherino.hasherino_dataclasses import Badge, Emote

__all__ = ["TwitchWebsocket", "ParsedMessage"]


class Command(Enum):
    PRIVMSG = auto()
    USERSTATE = auto()
    GLOBALUSERSTATE = auto()
    OTHER = auto()


class ParsedMessage:
    def __init__(self, message: str):
        self.source = self.tags = self.parameters = self.command = None

        raw_components = self._get_raw_components(message)
        if not raw_components:
            return None

        self.command = self._parse_command(raw_components["raw_command"])

        if self.command is None:
            return None
        else:
            if raw_components["raw_tags"] is not None:
                self.tags = self._parse_tags(raw_components["raw_tags"])

            self.source = self._parse_source(raw_components["raw_source"])
            self.parameters = raw_components["raw_parameters"]

    def get_badges(self, ttv_badges: dict) -> list[Badge]:
        def get_badge(set_id: str, version: str) -> dict | None:
            try:
                id_match = next((s for s in ttv_badges if s["set_id"] == set_id))
                version_match = next(
                    (s for s in id_match["versions"] if s["id"] == version)
                )
                return version_match
            except:
                return None

        badges = []

        if not self.tags or not self.tags.get("badges"):
            return badges

        try:
            for id, version in self.tags["badges"].items():
                badge = get_badge(id, version)
                if badge:
                    badges.append(Badge(id, badge["title"], badge["image_url_4x"]))
        except Exception as e:
            logging.exception(f"Error {e}. Failed to get badges from message: {self}")
            return []

        return badges

    def get_author_chat_color(self) -> str:
        result = "ffffff"

        if self.tags and self.tags["color"]:
            result = (
                self.tags["color"][1:]
                if self.tags["color"][0] == "#"
                else self.tags["color"]
            )

        assert len(result) <= 7, f"Returned invalid color: {result}"
        assert result[0] != "#"

        return f"#{result}"

    def get_author_displayname(self) -> str:
        if not self.tags or not self.tags.get("display-name"):
            logging.warning(f"Failed to get author display-name fo message: {self}")
            return ""

        return self.tags.get("display-name")

    def get_command(self) -> Command:
        result = Command.OTHER

        if not self.command or not self.command.get("command"):
            return result

        match self.command["command"]:
            case "PRIVMSG":
                result = Command.PRIVMSG
            case "USERSTATE":
                result = Command.USERSTATE
            case "GLOBALUSERSTATE":
                result = Command.GLOBALUSERSTATE
            case _:
                result = Command.OTHER

        return result

    def is_me(self) -> bool:
        """
        Messages sent with /me, coloring the whole line with the user's chat color
        """
        return (
            self.get_command() is Command.PRIVMSG
            and self.parameters[:7] == "\x01ACTION"
        )

    def get_message_text(self) -> str:
        # Remove \r\n from end of text
        result = "" if len(self.parameters) <= 2 else self.parameters[:-2]

        if self.is_me():
            # parameters: '\x01ACTION asd\x01\r\n'
            result = result[8:-1]

        return result

    def get_emote_sets(self) -> list[str]:
        result = []

        if self.tags and self.tags.get("emote-sets"):
            result = [tag for tag in self.tags["emote-sets"].split(",")]

        return result

    def get_message_elements(
        self, stv_emotes: dict[str, str] | None
    ) -> list[str | Emote]:
        if not self.tags:
            return []

        message_elements: list[str | Emote] = []
        emote_name_to_id_and_url: dict[str, tuple[str, str]] = {}

        if self.tags.get("emotes"):
            for emote_id, list_of_index_tuples in self.tags["emotes"].items():
                first_starting_index, first_ending_index = map(
                    int, list_of_index_tuples[0]
                )
                emote_name = self.get_message_text()[
                    first_starting_index : first_ending_index + 1
                ]
                emote_name_to_id_and_url[emote_name] = (
                    emote_id,
                    f"https://static-cdn.jtvnw.net/emoticons/v2/{emote_id}/default/dark/2.0",
                )

        if stv_emotes:
            for emote_name, emote_id in stv_emotes.items():
                emote_name_to_id_and_url[emote_name] = (
                    emote_id,
                    f"https://cdn.7tv.app/emote/{emote_id}/2x.webp",
                )

        for text in self.get_message_text().split(" "):
            is_emote = text in emote_name_to_id_and_url

            if is_emote:
                emote_id, url = emote_name_to_id_and_url[text]
                element = Emote(text, emote_id, url)
            else:
                element = text

            message_elements.append(element)

        return message_elements

    def __str__(self) -> str:
        return str(self.__dict__)

    def _get_raw_components(self, message: str) -> dict[str, str]:
        if not message:
            return None

        raw_tags = raw_source = raw_command = raw_parameters = ""

        # Start index
        idx = 0

        # Get tags
        if message[idx] == "@":
            end_idx = message.find(" ")
            raw_tags = message[1:end_idx]
            idx = end_idx + 1

        # Get source(nick and host)
        if message[idx] == ":":
            idx += 1
            end_idx = message.find(" ", idx)
            raw_source = message[idx:end_idx]
            idx = end_idx + 1

        # Command
        end_idx = message.find(":", idx)
        if -1 == end_idx:
            end_idx = len(message)

        raw_command = message[idx:end_idx].strip()

        # Parameters
        if end_idx != len(message):
            idx = end_idx + 1
            raw_parameters = message[idx:]

        return {
            "raw_tags": raw_tags,
            "raw_source": raw_source,
            "raw_command": raw_command,
            "raw_parameters": raw_parameters,
        }

    def _parse_command(self, raw_command: str) -> dict:
        parsed_command = None
        command_parts = raw_command.split(" ")

        match command_parts[0]:
            case "JOIN" | "PART" | "NOTICE" | "CLEARCHAT" | "HOSTTARGET":
                pass
            case "PRIVMSG":
                parsed_command = {
                    "command": command_parts[0],
                    "channel": command_parts[1],
                }
            case "PING":
                parsed_command = {"command": command_parts[0]}
            case "CAP":
                """
                The parameters part of the messages contains the
                enabled capabilities.
                """
                parsed_command = {
                    "command": command_parts[0],
                    "isCapRequestEnabled": command_parts[2] == "ACK",
                }
            case "GLOBALUSERSTATE":
                """
                Included only if you request the /commands capability.
                But it has no meaning without also including the /tags capability.
                """
                parsed_command = {"command": command_parts[0]}
            case "USERSTATE":
                """
                Included only if you request the /commands capability.
                But it has no meaning without also including the /tags capability.
                """
                parsed_command = {"command": command_parts[0]}
            case "ROOMSTATE":
                """
                Included only if you request the /commands capability.
                But it has no meaning without also including the /tags capability.
                """
                parsed_command = {
                    "command": command_parts[0],
                    "channel": command_parts[1],
                }
            case "RECONNECT":
                logging.info(
                    "The Twitch IRC server is about to terminate the connection for maintenance."
                )
                parsed_command = {"command": command_parts[0]}
            case "421":
                logging.warning(f"Unsupported IRC command: {command_parts[2]}")
                return None
            case "001":
                # Logged in (successfully authenticated)
                parsed_command = {
                    "command": command_parts[0],
                    "channel": command_parts[1],
                }
            case "002" | "003" | "004" | "353" | "366" | "372" | "375":
                """
                Ignoring all other numeric messages.
                353 tells you who else is in the chat room you're joining.
                """
                pass
            case "376":
                logging.info(f"Numeric message: {command_parts[0]}")
                return None
            case _:
                logging.warning(f"Unexpected command: {command_parts[0]}")
                return None

        return parsed_command

    def _parse_source(self, raw_source: str) -> None | dict[str, str]:
        if not raw_source:
            return None
        else:
            source_parts = raw_source.split("!")
            return {
                "nick": source_parts[0] if len(source_parts) == 2 else None,
                "host": source_parts[1] if len(source_parts) == 2 else source_parts[0],
            }

    def _parse_tags(self, raw_tags: str):
        dict_parsed_tags = {}

        if not raw_tags:
            return dict_parsed_tags

        for tag in raw_tags.split(";"):
            tag_key, tag_value = tag.split("=")

            match tag_key:
                case "badges-info":
                    """
                    Contains metadata related to the chat badges in the badges tag.
                    Currently, this tag contains metadata only for subscriber badges, to indicate the number of months the user has been a subscriber.
                    """
                    pass
                case "badges":
                    # badges=staff/1,broadcaster/1,turbo/1;
                    if tag_value:
                        badges = dict()
                        for badge_and_version in tag_value.split(","):
                            badge, version = badge_and_version.split("/")
                            badges[badge] = version
                    else:
                        badges = None

                    dict_parsed_tags[tag_key] = badges

                case "emotes":
                    """
                    emotes=25:0-4,12-16/1902:6-10
                    emotes=emotesv2_c51307f86f6241bc8cd8385efd7c7509:0-9/emotesv2_d9f1e820ca8e42bab70fc2f22dea0d5a:31-44

                    Comma-delimited list of emotes and their positions in the message.
                    Each emote is in the form, <emote ID>:<start position>-<end position>
                    """
                    if tag_value:
                        id_to_positions = defaultdict(list)

                        for emote_id_and_pos in tag_value.split("/"):
                            emote_id, positions = emote_id_and_pos.split(":")

                            for start_end in positions.split(","):
                                start, end = start_end.split("-")

                                id_to_positions[emote_id].append((start, end))

                        dict_parsed_tags[tag_key] = dict(id_to_positions)
                    else:
                        dict_parsed_tags[tag_key] = None

                case "color":
                    dict_parsed_tags["color"] = tag_value[1:]

                case "user-id":
                    dict_parsed_tags["user-id"] = tag_value

                case "display-name":
                    dict_parsed_tags["display-name"] = tag_value

                case "emote-sets":
                    dict_parsed_tags["emote-sets"] = tag_value

                case _:
                    pass

        return dict_parsed_tags


class TwitchWebsocket:
    def __init__(self) -> None:
        self._websocket = None

    async def is_connected(self) -> bool:
        return self._websocket is not None

    async def _authenticate(self, token: str, user: str):
        """
        Returns parsed authentication response
        """
        if self._websocket is None:
            raise Exception("Websocket not connected")

        await self._websocket.send(f"PASS oauth:{token}")
        await self._websocket.send(f"NICK {user}")
        await self._websocket.send(f"USER {user} 8 * :{user}")

    async def join_channel(self, channel: str):
        if self._websocket is None:
            raise Exception("Websocket not connected")

        await self._websocket.send(f"JOIN #{channel}")

    async def leave_channel(self, channel: str):
        if self._websocket is None:
            raise Exception("Websocket not connected")

        await self._websocket.send(f"PART #{channel}")

    async def send_message(self, channel: str, message: str):
        logging.debug(f"Sending message on channel {channel} message: {message}")

        if self._websocket is None:
            raise Exception("Websocket not connected")

        logging.debug("Acquired lock, sending message")
        await self._websocket.send(f"PRIVMSG #{channel} :{message}")

    async def listen_message(
        self,
        message_callback: Awaitable,
        reconnect_callback: Awaitable[bool],
        token: str,
        username: str,
        join_channel: str | None = None,
    ):
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        async for websocket in websockets.connect(
            "wss://irc-ws.chat.twitch.tv:443",
            ping_interval=3,
            ping_timeout=2,
            ssl=ssl_context,
        ):
            try:
                self._websocket = websocket

                await websocket.send(
                    "CAP REQ :twitch.tv/commands twitch.tv/membership twitch.tv/tags"
                )
                await self._authenticate(token, username)
                await reconnect_callback(False)

                if join_channel:
                    await self.join_channel(join_channel)

                async for message in websocket:
                    parsed_message: ParsedMessage = ParsedMessage(message)
                    await message_callback(parsed_message)

            except ConnectionClosedError:
                logging.warning("Websocket connection closed, reconnecting")
                self._websocket = None
                await reconnect_callback(True)

            except Exception as e:
                logging.exception(f"Websocket connection failed: {e}")



import asyncio
import json
import logging
import sys
from abc import ABC
from io import TextIOBase
from pathlib import Path
from typing import Any

import keyring
from flet import Page


def get_default_os_settings_path() -> Path:
    """
    Returns the default path where hasherino files are stored according on the user's OS.

    Windows: %APPDATA%/hasherino
    Linux: $HOME/.local/share/hasherino
    Others: same folder the app is running from
    """
    home = Path.home()

    if sys.platform.startswith("win32"):
        return home / "AppData" / "hasherino"
    elif sys.platform.startswith("linux"):
        return home / ".local" / "share" / "hasherino"
    else:
        return Path.cwd()


class AsyncKeyValueStorage(ABC):
    async def get(self, key) -> Any:
        pass

    async def set(self, key, value):
        pass

    async def remove(self, key):
        pass


class MemoryOnlyStorage(AsyncKeyValueStorage):
    def __init__(self, page: Page) -> None:
        super().__init__()
        self.page = page

    async def get(self, key) -> Any:
        value = self.page.session.get(key)
        return value

    async def set(self, key, value):
        logging.debug(f"Memory storage set {key} to {value}")
        self.page.session.set(key, value)

    async def remove(self, key):
        logging.debug(f"Memory storage removed {key}")
        self.page.session.remove(key)


class PersistentStorage(AsyncKeyValueStorage):
    """
    Persistent async key-value storage.

    Locks implemented based on wikipedia's pseucode for a Readers–writer lock
    """

    def __init__(self, file: TextIOBase | str = "db.json") -> None:
        """
        File can be the file name string to a database file or a TextIOBase if you don't want to use a file,
        such as using a StringIO object for a memory database
        """
        self._file = file

        self._r = asyncio.Lock()
        self._g = asyncio.Lock()
        self._b = 0

        with self.get_file() as file_object:
            try:
                self._data: dict = json.load(file_object)
            except json.JSONDecodeError:
                # File is not empty, database exists but failed to load
                if file_object.read():
                    raise Exception("Failed to load database.")

                # File is empty so it's a new database, make an empty dict
                self._data: dict = {}

            assert type(self._data) == dict, "Database file is not a dictionary"

    def get_file(self):
        if type(self._file) == str:
            fpath = get_default_os_settings_path() / self._file
            return open(fpath, "r+" if fpath.is_file() else "w+")
        elif isinstance(self._file, TextIOBase):
            return self._file
        else:
            raise Exception("Invalid file type")

    async def _begin_read(self):
        await self._r.acquire()
        self._b += 1
        if self._b == 1:
            await self._g.acquire()
        self._r.release()

    async def _end_read(self):
        await self._r.acquire()
        self._b -= 1
        if self._b == 0:
            self._g.release()
        self._r.release()

    async def _begin_write(self):
        await self._g.acquire()

    async def _end_write(self):
        self._g.release()

    async def get(self, key) -> Any:
        if key == "token":
            return keyring.get_password("hasherino", "token")

        await self._begin_read()

        result = self._data.get(key, None)

        await self._end_read()

        return result

    async def set(self, key, value):
        if key == "token":
            keyring.set_password("hasherino", "token", value)
            # DO NOT log passwords
            return

        await self._begin_write()

        logging.debug(f"Persistent storage set {key} to {value}")

        self._data[key] = value
        with self.get_file() as file_object:
            file_object.truncate(0)
            file_object.seek(0)
            json.dump(self._data, file_object, sort_keys=True, indent=4)

        await self._end_write()

    async def remove(self, key):
        await self._begin_write()

        logging.debug(f"Persistent storage removed {key}")
        self._data.pop(key)
        with self.get_file() as file_object:
            file_object.truncate(0)
            file_object.seek(0)
            json.dump(self._data, file_object, sort_keys=True, indent=4)

        await self._end_write()



import asyncio
import logging

import flet as ft

from hasherino import user_auth
from hasherino.api import helix
from hasherino.components import (
    AccountDialog,
    ChatContainer,
    NewMessageRow,
    SettingsView,
    StatusColumn,
    Tabs,
)
from hasherino.components.settings_view import LOG_PATH
from hasherino.hasherino_dataclasses import Emote, HasherinoUser, Message
from hasherino.pubsub import PubSub
from hasherino.storage import (
    AsyncKeyValueStorage,
    MemoryOnlyStorage,
    PersistentStorage,
    get_default_os_settings_path,
)
from hasherino.twitch_websocket import Command, ParsedMessage, TwitchWebsocket


class Hasherino:
    def __init__(
        self,
        font_size_pubsub: PubSub,
        memory_storage: AsyncKeyValueStorage,
        persistent_storage: AsyncKeyValueStorage,
        page: ft.Page,
    ) -> None:
        self.font_size_pubsub = font_size_pubsub
        self.memory_storage = memory_storage
        self.persistent_storage = persistent_storage
        self.page = page
        self.page.is_ctrl_pressed = False
        self.message_listener: None | asyncio.Task = None
        self.emote_set_cache: dict[str, list[Emote]] = dict()

    async def login_click(self, _):
        logging.debug("Clicked login")

        app_id = await self.persistent_storage.get("app_id")
        token = await user_auth.request_oauth_token(app_id)
        users = await helix.get_users(app_id, token, [])

        if users:
            websocket: TwitchWebsocket = await self.memory_storage.get("websocket")

            if self.message_listener:
                self.message_listener.cancel()
                self.message_listener = None

            self.message_listener = asyncio.create_task(
                websocket.listen_message(
                    message_callback=self.message_received,
                    reconnect_callback=self.status_column.set_reconnecting_status,
                    token=token,
                    username=users[0].login,
                    join_channel=await self.persistent_storage.get("channel"),
                )
            )

            asyncio.gather(
                self.persistent_storage.set("token", token),
                self.persistent_storage.set("user_name", users[0].display_name),
                self.persistent_storage.set("user_id", users[0].id),
                self.memory_storage.set(
                    "ttv_badges", await helix.get_global_badges(app_id, token)
                ),
            )
        else:
            self.page.dialog = ft.AlertDialog(
                content=ft.Text("Failed to authenticate.")
            )
            self.page.dialog.open = True

        await self.page.update_async()

    async def settings_click(self, _):
        logging.debug("Clicked on settings")
        sv = SettingsView(self.font_size_pubsub, self.persistent_storage)
        await sv.init()
        self.page.views.append(sv)
        await self.page.update_async()

    async def message_received(self, message: ParsedMessage):
        logging.debug(f"Received message with command {message.get_command()}")

        match message.get_command():
            case Command.USERSTATE:
                if (
                    message.get_author_displayname().lower()
                    == (await self.persistent_storage.get("user_name")).lower()
                ):
                    async with asyncio.TaskGroup() as tg:
                        tg.create_task(
                            self.memory_storage.set(
                                "user_badges",
                                message.get_badges(
                                    await self.memory_storage.get("ttv_badges")
                                ),
                            )
                        )

                        tg.create_task(
                            self.memory_storage.set(
                                "user_color", message.get_author_chat_color()
                            )
                        )

                        if not await self.memory_storage.get("ttv_emote_sets"):
                            emotes: dict[str, Emote] = dict()

                            for emote_obj in await helix.get_emote_sets(
                                await self.persistent_storage.get("app_id"),
                                await self.persistent_storage.get("token"),
                                message.get_emote_sets(),
                            ):
                                emotes[emote_obj["name"]] = Emote(
                                    name=emote_obj["name"],
                                    id=emote_obj["id"],
                                    url=f"https://static-cdn.jtvnw.net/emoticons/v2/{emote_obj['id']}/default/dark/2.0",
                                )

                            tg.create_task(
                                self.memory_storage.set("ttv_emote_sets", emotes)
                            )

            case Command.PRIVMSG:
                author: str = message.get_author_displayname()
                stv_emotes = await self.memory_storage.get("7tv_emotes")
                channel = await self.persistent_storage.get("channel")
                channel_emotes = (
                    stv_emotes[channel] if stv_emotes and channel in stv_emotes else {}
                )

                await self.chat_container_on_msg(
                    Message(
                        HasherinoUser(
                            name=author,
                            badges=message.get_badges(
                                await self.memory_storage.get("ttv_badges")
                            ),
                            chat_color=message.get_author_chat_color(),
                        ),
                        elements=message.get_message_elements(channel_emotes),
                        message_type="chat_message",
                        me=message.is_me(),
                    )
                )
            case _:
                pass

    async def select_chat_click(self, _):
        channel = ft.TextField(label="Channel", autofocus=True)
        logging.debug("Clicked on select chat")

        async def join_chat_click(_):
            websocket: TwitchWebsocket = await self.memory_storage.get("websocket")
            channel.error_text = ""

            if await self.persistent_storage.get("channel"):
                logging.info(
                    f"Leaving channel {await self.persistent_storage.get('channel')}"
                )
                await websocket.leave_channel(
                    await self.persistent_storage.get("channel")
                )

            logging.info(f"Joining channel {channel.value}")

            try:
                await websocket.join_channel(channel.value)
            except Exception:
                channel.error_text = f"Not logged in"
                logging.error(channel.error_text)
                await self.page.update_async()
                return

            await self.tabs.add_tab(channel.value)
            await self.persistent_storage.set("channel", channel.value)
            self.page.dialog.open = False

            await self.page.update_async()

        channel.on_submit = join_chat_click

        self.page.dialog = ft.AlertDialog(
            content=channel,
            actions=[ft.ElevatedButton(text="Join", on_click=join_chat_click)],
        )
        self.page.dialog.open = True
        await self.page.update_async()

    async def on_resize(self, _):
        if self.page.window_height > 100 and self.page.window_width > 100:
            async with asyncio.TaskGroup() as tg:
                tg.create_task(
                    self.persistent_storage.set(
                        "window_height", self.page.window_height
                    )
                )
                tg.create_task(
                    self.persistent_storage.set("window_width", self.page.window_width)
                )

    async def on_kb_event(self, e: ft.KeyboardEvent):
        self.page.is_ctrl_pressed = e.ctrl

        if e.key == "E" and e.ctrl:
            await self.new_message_row.emote_completion()

        elif e.key == "U" and e.ctrl:
            await self.new_message_row.user_completion()

    async def run(self):
        self.page.window_width = await self.persistent_storage.get("window_width")
        self.page.window_height = await self.persistent_storage.get("window_height")
        self.page.on_keyboard_event = self.on_kb_event

        match await self.persistent_storage.get("theme"):
            case "System":
                self.page.theme_mode = ft.ThemeMode.SYSTEM
            case "Dark mode":
                self.page.theme_mode = ft.ThemeMode.DARK
            case "Light mode":
                self.page.theme_mode = ft.ThemeMode.LIGHT
            case _:
                pass

        self.page.on_resize = self.on_resize
        self.page.horizontal_alignment = "stretch"
        self.page.title = "Hasherino"

        self.page.dialog = AccountDialog(self.persistent_storage)
        self.status_column = StatusColumn(self.memory_storage, self.persistent_storage)
        chat_container = ChatContainer(
            self.persistent_storage, self.memory_storage, self.font_size_pubsub
        )
        self.new_message_row = NewMessageRow(
            self.memory_storage,
            self.persistent_storage,
            chat_container.on_message,
            self.status_column.set_reconnecting_status,
        )
        self.tabs = Tabs(self.memory_storage, self.persistent_storage)

        self.chat_container_on_msg = chat_container.on_message

        # Add everything to the page
        await self.page.add_async(
            ft.Row(
                [
                    self.tabs,
                    ft.Row(
                        controls=[
                            ft.IconButton(
                                icon=ft.icons.LOGIN, on_click=self.login_click
                            ),
                            ft.IconButton(
                                icon=ft.icons.CHAT, on_click=self.select_chat_click
                            ),
                            ft.IconButton(
                                icon=ft.icons.SETTINGS, on_click=self.settings_click
                            ),
                        ]
                    ),
                ],
                alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
            ),
            chat_container,
            self.new_message_row,
            self.status_column,
        )

        if user_name := await self.persistent_storage.get("user_name"):
            websocket: TwitchWebsocket = await self.memory_storage.get("websocket")

            channel = await self.persistent_storage.get("channel")
            token = await self.persistent_storage.get("token")

            self.message_listener = asyncio.create_task(
                websocket.listen_message(
                    message_callback=self.message_received,
                    reconnect_callback=self.status_column.set_reconnecting_status,
                    token=token,
                    username=user_name,
                    join_channel=channel,
                )
            )

            if channel:
                await self.tabs.add_tab(channel)

            await self.memory_storage.set(
                "ttv_badges",
                await helix.get_global_badges(
                    await self.persistent_storage.get("app_id"),
                    token,
                ),
            )


async def main(page: ft.Page):
    # Make settings folder for user's OS if it doesn't exist
    if not get_default_os_settings_path().exists():
        # Uses print cause logging isn't set up at this point
        print(f"Making settings directory on {get_default_os_settings_path()}")
        get_default_os_settings_path().mkdir(parents=True)

    logging.basicConfig(
        level=logging.DEBUG,
        format="%(asctime)s | %(name)s | %(filename)s | %(levelname)s | %(funcName)s | %(lineno)d | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        handlers=[
            logging.StreamHandler(),  # Outputs logs to the console
            logging.FileHandler(LOG_PATH, mode="w"),
        ],
    )

    websockets_logger = logging.getLogger("websockets")
    websockets_logger.setLevel(logging.INFO)

    persistent_storage = PersistentStorage()
    memory_storage = MemoryOnlyStorage(page)

    app_id = "hvmj7blkwy2gw3xf820n47i85g4sub"

    websocket = TwitchWebsocket()
    await memory_storage.set("websocket", websocket)

    if await persistent_storage.get("token"):
        renewed_token = await user_auth.request_oauth_token(
            app_id, await persistent_storage.get("token")
        )
        await persistent_storage.set("token", renewed_token)

    if not await persistent_storage.get("not_first_run"):
        async with asyncio.TaskGroup() as tg:
            tg.create_task(persistent_storage.set("app_id", app_id))
            tg.create_task(persistent_storage.set("chat_font_size", 18))
            tg.create_task(persistent_storage.set("chat_update_rate", 0.5))
            tg.create_task(persistent_storage.set("color_switcher", False))
            tg.create_task(persistent_storage.set("max_messages_per_chat", 100))
            tg.create_task(persistent_storage.set("not_first_run", True))
            tg.create_task(persistent_storage.set("theme", "System"))
            tg.create_task(persistent_storage.set("window_width", 500))
            tg.create_task(persistent_storage.set("window_height", 800))

    hasherino = Hasherino(PubSub(), memory_storage, persistent_storage, page)
    await hasherino.run()


def run_hasherino():
    # Script entrypoint
    ft.app(target=main)


if __name__ == "__main__":
    run_hasherino()



import flet as ft

from hasherino.storage import AsyncKeyValueStorage


class StatusColumn(ft.Column):
    def __init__(
        self,
        memory_storage: AsyncKeyValueStorage,
        persistent_storage: AsyncKeyValueStorage,
    ):
        self.reconnecting_status = ft.Row(
            vertical_alignment=ft.CrossAxisAlignment.CENTER,
            controls=[
                ft.ProgressRing(width=16, height=16, stroke_width=2),
                ft.Text("Reconnecting..."),
            ],
        )

        self.memory_storage = memory_storage
        self.persistent_storage = persistent_storage

        super().__init__()

    async def set_reconnecting_status(self, reconnecting: bool):
        await self.memory_storage.set("reconnecting", reconnecting)

        if reconnecting:
            self.controls.append(self.reconnecting_status)
        else:
            if self.reconnecting_status in self.controls:
                self.controls.remove(self.reconnecting_status)

            channel = await self.persistent_storage.get("channel")
            if channel:
                websocket = await self.memory_storage.get("websocket")
                await websocket.join_channel(channel)

        await self.page.update_async()



from hasherino.components.account_dialog import AccountDialog
from hasherino.components.chat_container import ChatContainer
from hasherino.components.chat_message import ChatMessage
from hasherino.components.new_message_row import NewMessageRow
from hasherino.components.settings_view import SettingsView
from hasherino.components.status_column import StatusColumn
from hasherino.components.tabs import Tabs



from typing import Any, Awaitable


class PubSub:
    def __init__(self) -> None:
        self.funcs: set[Awaitable] = set()

    async def subscribe(self, func: Awaitable):
        self.funcs.add(func)

    async def subscribe_all(self, funcs: list[Awaitable]):
        self.funcs.update(funcs)

    async def send(self, message: Any):
        for func in self.funcs:
            await func(message)



from abc import ABC

import flet as ft

from hasherino.hasherino_dataclasses import Emote, Message
from hasherino.pubsub import PubSub


class FontSizeSubscriber(ABC):
    async def on_font_size_changed(self, new_font_size: int):
        ...


class ChatText(ft.Text, FontSizeSubscriber):
    def __init__(self, text: str, color: str, size: int, weight=""):
        super().__init__(text, size=size, weight=weight, color=color, selectable=True)

    async def on_font_size_changed(self, new_font_size: int):
        self.size = new_font_size


class ChatBadge(ft.Image, FontSizeSubscriber):
    def __init__(self, src: str, height: int):
        super().__init__(src=src, height=height)

    async def on_font_size_changed(self, new_font_size: int):
        self.height = new_font_size


class ChatEmote(ft.Image, FontSizeSubscriber):
    async def on_font_size_changed(self, new_font_size: int):
        self.height = new_font_size * 2


class ChatMessage(ft.Row):
    def __init__(self, message: Message, page: ft.Page, font_size: int):
        super().__init__()
        self.vertical_alignment = "start"
        self.wrap = True
        self.width = page.width
        self.page = page
        self.font_size = font_size
        self.spacing = 2
        self.run_spacing = 0
        self.vertical_alignment = ft.CrossAxisAlignment.CENTER

        self.add_control_elements(message)

    def add_control_elements(self, message):
        self.controls = [
            ChatBadge(badge.url, self.font_size) for badge in message.user.badges
        ]

        self.controls.append(
            ChatText(
                f"{message.user.name}: ",
                message.user.chat_color,
                self.font_size,
                weight="bold",
            )
        )

        for element in message.elements:
            if type(element) == str:
                color = message.user.chat_color if message.me else ""
                result = ChatText(element, color, self.font_size)
            elif type(element) == Emote:
                result = ChatEmote(
                    src=element.url,
                    height=self.font_size * 2,
                )
            else:
                raise TypeError

            self.controls.append(result)

    async def subscribe_to_font_size_change(self, pubsub: PubSub):
        await pubsub.subscribe_all(
            [
                control.on_font_size_changed
                for control in self.controls
                if isinstance(control, FontSizeSubscriber)
            ]
        )



import asyncio
import enum
import logging
import ssl
from dataclasses import dataclass
from typing import Iterable

import certifi
from aiohttp import ClientSession, TCPConnector

__all__ = ["get_users", "update_chat_color", "TwitchUser", "NormalUserColor"]

_BASE_URL = "https://api.twitch.tv/helix/"


@dataclass
class TwitchUser:
    id: str
    login: str
    display_name: str
    type: str  # admin, global_mod, staff, or "" for normal user
    broadcaster_type: str  # affiliate, partner or "" for normal broadcaster
    description: str
    profile_image_url: str
    offline_image_url: str
    created_at: str

    def __init__(self, json_content: dict):
        self.__dict__ = json_content


@dataclass
class UserChatColor:
    user_id: str
    user_login: str
    user_name: str
    color: str

    def __init__(self, json_content: dict):
        self.__dict__ = json_content


class NormalUserColor(enum.StrEnum):
    blue = "#0000ff"
    blue_violet = "#8a2be2"
    cadet_blue = "#5f9ea0"
    chocolate = "#d2691e"
    coral = "#ff7f50"
    dodger_blue = "#1e90ff"
    firebrick = "#b22222"
    golden_rod = "#daa520"
    green = "#008000"
    hot_pink = "#ff69b4"
    orange_red = "#ff4500"
    red = "#ff0000"
    sea_green = "#2e8b57"
    spring_green = "#00ff7f"
    yellow_green = "#9acd32"


async def get_users(
    app_id: str,
    oauth_token: str,
    users: Iterable[str | int],
) -> list[TwitchUser]:
    """
    Raises Exception for invalid status code or KeyError if the json response is invalid
    """
    users = "&".join(
        f"id={user}" if type(user) == int else f"login={user}" for user in users
    )
    logging.debug(f"Generated helix get user parameters: {users}")

    ssl_context = ssl.create_default_context(cafile=certifi.where())
    conn = TCPConnector(ssl=ssl_context)

    async with ClientSession(connector=conn) as session:
        async with session.get(
            f"{_BASE_URL}users",
            headers={
                "Authorization": f"Bearer {oauth_token}",
                "Client-Id": app_id,
            },
            params=users,
        ) as response:
            json_result = await response.json()
            logging.debug(f"Helix get user response: {json_result}")

            if response.status != 200:
                raise Exception("Unable to get user")

            return [TwitchUser(user) for user in json_result["data"]]


async def get_user_chat_color(
    app_id: str,
    oauth_token: str,
    user_ids: Iterable[int],
) -> list[UserChatColor]:
    """
    Raises Exception for invalid status code or KeyError if the json response is invalid
    """
    ids = "&".join(f"user_id={user_id}" for user_id in user_ids)
    logging.debug(f"Generated helix get user id parameters: {ids}")

    ssl_context = ssl.create_default_context(cafile=certifi.where())
    conn = TCPConnector(ssl=ssl_context)

    async with ClientSession(connector=conn) as session:
        async with session.get(
            f"{_BASE_URL}chat/color",
            headers={
                "Authorization": f"Bearer {oauth_token}",
                "Client-Id": app_id,
            },
            params=ids,
        ) as response:
            json_result = await response.json()
            logging.debug(f"Helix get user response: {json_result}")

            if response.status != 200:
                raise Exception("Unable to get user chat color")

            return [UserChatColor(user) for user in json_result["data"]]


async def update_chat_color(
    app_id: str, oauth_token: str, user_id: str, color_code: str | NormalUserColor
) -> bool:
    ssl_context = ssl.create_default_context(cafile=certifi.where())
    conn = TCPConnector(ssl=ssl_context)

    async with ClientSession(connector=conn) as session:
        params = {
            "user_id": user_id,
            "color": str(color_code),
        }
        async with session.put(
            f"{_BASE_URL}chat/color",
            headers={
                "Authorization": f"Bearer {oauth_token}",
                "Client-Id": app_id,
            },
            params=params,
        ) as response:
            logging.debug(
                f"Helix tried changing user color, response code: {response.status}. Params: {params}"
            )
            return response.status == 204


async def get_global_badges(
    app_id: str,
    oauth_token: str,
) -> dict:
    """
    Raises Exception for invalid status code
    """
    ssl_context = ssl.create_default_context(cafile=certifi.where())
    conn = TCPConnector(ssl=ssl_context)

    async with ClientSession(connector=conn) as session:
        async with session.get(
            f"{_BASE_URL}chat/badges/global",
            headers={
                "Authorization": f"Bearer {oauth_token}",
                "Client-Id": app_id,
            },
        ) as response:
            json_result = await response.json()
            logging.debug(f"Helix get global badge response: {json_result}")

            if response.status != 200:
                raise Exception("Unable to get global badges")

            return json_result["data"]


async def get_channel_emotes(app_id: str, oauth_token: str, broadcaster_id: str):
    """
    Raises Exception for invalid status code
    """
    ssl_context = ssl.create_default_context(cafile=certifi.where())
    conn = TCPConnector(ssl=ssl_context)

    async with ClientSession(connector=conn) as session:
        async with session.get(
            f"{_BASE_URL}chat/emotes",
            headers={
                "Authorization": f"Bearer {oauth_token}",
                "Client-Id": app_id,
            },
            params=f"broadcaster_id={broadcaster_id}",
        ) as response:
            json_result = await response.json()
            logging.debug(f"Helix get channel emotes response: {json_result}")

            if not response.ok:
                raise Exception(
                    f"Unable to get channel emotes for {broadcaster_id} with response {json_result}"
                )

            return json_result["data"]


async def get_emote_sets(app_id: str, oauth_token: str, emote_set_ids: list[str]):
    """
    Raises Exception for invalid status code
    """
    if not emote_set_ids:
        return []

    if len(emote_set_ids) > 25:
        raise Exception("You may specify a maximum of 25 IDs.")

    ssl_context = ssl.create_default_context(cafile=certifi.where())
    conn = TCPConnector(ssl=ssl_context)

    ids = "&".join(f"emote_set_id={user_id}" for user_id in emote_set_ids)
    logging.debug(f"Generated params for emote sets query: {ids}")

    async with ClientSession(connector=conn) as session:
        async with session.get(
            f"{_BASE_URL}chat/emotes/set",
            headers={
                "Authorization": f"Bearer {oauth_token}",
                "Client-Id": app_id,
            },
            params=ids,
        ) as response:
            json_result = await response.json()
            logging.debug(f"Helix get channel emotes response: {json_result}")

            if not response.ok:
                raise Exception(
                    f"Unable to get emote sets {','.join(emote_set_ids)} with response {json_result}"
                )

            return json_result["data"]



import flet as ft

from hasherino.storage import AsyncKeyValueStorage


class AccountDialog(ft.AlertDialog):
    def __init__(self, storage: AsyncKeyValueStorage):
        # A dialog asking for a user display name
        self.join_user_name = ft.TextField(
            label="Enter your name to join the chat",
            autofocus=True,
            on_submit=self.join_chat_click,
        )
        self.storage = storage
        super().__init__(
            open=True,
            modal=True,
            title=ft.Text("Welcome!"),
            content=ft.Column([self.join_user_name], width=300, height=70, tight=True),
            actions=[
                ft.ElevatedButton(text="Join chat", on_click=self.join_chat_click)
            ],
            actions_alignment="end",
        )

    async def join_chat_click(self, _):
        if not self.join_user_name.value:
            self.join_user_name.error_text = "Name cannot be blank!"
            await self.join_user_name.update_async()
        else:
            await self.storage.set("user_name", self.join_user_name.value)
            self.page.dialog.open = False
            await self.page.update_async()



import asyncio
import logging
from random import choice
from typing import Awaitable

import flet as ft

from hasherino.api import helix
from hasherino.api.helix import NormalUserColor
from hasherino.hasherino_dataclasses import Emote, HasherinoUser, Message
from hasherino.storage import AsyncKeyValueStorage


class NewMessageRow(ft.Row):
    def __init__(
        self,
        memory_storage: AsyncKeyValueStorage,
        persistent_storage: AsyncKeyValueStorage,
        chat_container_on_message: Awaitable,
        reconnect_callback: Awaitable,
    ):
        self.memory_storage = memory_storage
        self.persistent_storage = persistent_storage
        self.chat_container_on_message = chat_container_on_message
        self.reconnect_callback = reconnect_callback

        # A new message entry form
        self.new_message = ft.TextField(
            hint_text="Write a message...",
            autofocus=True,
            shift_enter=True,
            min_lines=1,
            max_lines=5,
            filled=True,
            expand=True,
            on_submit=self.send_message_click,
            on_focus=self.new_message_focus,
            on_blur=self.new_message_clear_error,
            on_change=self.new_message_clear_error,
        )
        self.send_message = ft.IconButton(
            icon=ft.icons.SEND_ROUNDED,
            tooltip="Send message",
            on_click=self.send_message_click,
        )

        super().__init__([self.new_message, self.send_message])

    async def new_message_clear_error(self, e):
        e.control.error_text = ""
        await self.page.update_async()

    async def emote_completion(self):
        if not self.new_message.value:
            return

        stv_emotes = await self.memory_storage.get("7tv_emotes")
        channel_stv_emotes = (
            stv_emotes[await self.persistent_storage.get("channel")]
            if stv_emotes
            else {}
        )
        emote_map: dict[str, Emote] = await self.memory_storage.get("ttv_emote_sets")
        emote_names = sorted(list(emote_map.keys()) + list(channel_stv_emotes.keys()))

        if emote_names:
            last_space_index = self.new_message.value.rfind(" ")

            # If a space is not found it'll return -1, so getting the last word with last_space_index+1 works either way.
            last_word = self.new_message.value[last_space_index + 1 :]

            logging.debug(
                f"Attempting emote completion. last_space_index: {last_space_index} last_word: {last_word}"
            )

            for emote_name in emote_names:
                if emote_name.lower().startswith(last_word.lower()):
                    self.new_message.value = (
                        f"{self.new_message.value[:-len(last_word)]}{emote_name}"
                    )
                    logging.debug(
                        f"Found emote completion for {last_word} -> {emote_name}."
                    )
                    await self.new_message.update_async()
                    return

    async def user_completion(self):
        user_list = await self.memory_storage.get("channel_user_list")

        if user_list and self.new_message.value:
            last_space_index = self.new_message.value.rfind(" ")

            # If a space is not found it'll return -1, so getting the last word with last_space_index+1 works either way.
            last_word = self.new_message.value[last_space_index + 1 :]

            logging.debug(
                f"Attempting username completion. last_space_index: {last_space_index} last_word: {last_word}"
            )

            sorted_user_list = sorted(
                user_list[await self.persistent_storage.get("channel")]
            )

            for user in sorted_user_list:
                if user.lower().startswith(last_word.lower()):
                    self.new_message.value = (
                        f"{self.new_message.value[:-len(last_word)]}{user}"
                    )
                    logging.debug(
                        f"Found username completion for {last_word} -> {user}."
                    )
                    await self.new_message.update_async()
                    return

    async def new_message_focus(self, e):
        if await self.persistent_storage.get("user_name"):
            e.control.prefix = ft.Text(
                f"{await self.persistent_storage.get('user_name')}: "
            )

            channel = await self.persistent_storage.get("channel")
            if channel:
                e.control.hint_text = f"Write a message on channel {channel}"
            else:
                e.control.hint_text = "Write a message..."

            await self.page.update_async()

    async def send_message_click(self, _):
        if self.new_message.value == "":
            return

        self.new_message.error_style = ft.TextStyle(size=16)

        disconnect_error = "Please connect to twitch before sending messages."

        websocket = await self.memory_storage.get("websocket")
        is_connected = websocket and await websocket.is_connected()
        if not is_connected:
            self.new_message.error_text = disconnect_error
            await self.update_async()
            return

        if not bool(await self.persistent_storage.get("user_name")):
            self.new_message.error_text = (
                "Please connect to twitch before sending messages."
            )
            await self.update_async()
            return

        if not await self.persistent_storage.get("channel"):
            self.new_message.error_text = (
                "Please connect to a channel before sending messages."
            )
            await self.update_async()
            return

        emote_map: dict[str, Emote] = await self.memory_storage.get("ttv_emote_sets")

        try:
            async with asyncio.timeout(2):
                await websocket.send_message(
                    await self.persistent_storage.get("channel"), self.new_message.value
                )
        except (asyncio.TimeoutError, Exception):
            await self.reconnect_callback(True)
            self.new_message.error_text = disconnect_error
            await self.update_async()
            return

        await self.chat_container_on_message(
            Message(
                HasherinoUser(
                    name=await self.persistent_storage.get("user_name"),
                    badges=await self.memory_storage.get("user_badges"),
                    chat_color=await self.memory_storage.get("user_color"),
                ),
                elements=[
                    emote_map[element] if element in emote_map else element
                    for element in self.new_message.value.split(" ")
                ],
                message_type="chat_message",
                me=False,
            )
        )

        if not self.page.is_ctrl_pressed:
            self.new_message.value = ""

        await self.new_message.focus_async()
        await self.page.update_async()

        if await self.persistent_storage.get("color_switcher"):
            # Using user_color can cause the color to repeat , since it gets replaced on USERSTATE messages
            color_index = await self.memory_storage.get("user_color_index")

            if not color_index:
                color_index = 0

            color_list = [str(c).upper() for c in list(NormalUserColor)]

            color_index = (color_index + 1) % len(color_list)

            try:
                next_color = color_list[color_index]
                logging.debug(
                    f"Next color index: {color_index}. List: {', '.join(color_list)}"
                )
                await self.memory_storage.set("user_color_index", color_index)
            except (ValueError, IndexError) as e:
                logging.error(f"Error trying to get next switcher chat color: {e}")
                next_color = choice(color_list)

            try:
                await helix.update_chat_color(
                    await self.persistent_storage.get("app_id"),
                    await self.persistent_storage.get("token"),
                    await self.persistent_storage.get("user_id"),
                    next_color,
                )
                logging.info(f"Switcher set user color to {next_color}")

            except Exception as e:
                logging.error(f"Switcher failed to switch user chat color: {e}")



import logging
from pathlib import Path

import flet as ft

from hasherino.pubsub import PubSub
from hasherino.storage import AsyncKeyValueStorage, get_default_os_settings_path

LOG_PATH = get_default_os_settings_path() / "hasherino.log"


class SettingsView(ft.View):
    def __init__(self, font_size_pubsub: PubSub, storage: AsyncKeyValueStorage):
        self.font_size_pubsub = font_size_pubsub
        self.storage = storage

    async def init(self):
        super().__init__(
            "/settings",
            [
                ft.IconButton(icon=ft.icons.ARROW_BACK, on_click=self._back_click),
                ft.Tabs(
                    tabs=[
                        await self._get_general_tab(),
                        await self._get_appearance_tab(),
                        await self._get_debug_tab(),
                    ]
                ),
            ],
        )

    async def _get_general_tab(self) -> ft.Tab:
        return ft.Tab(
            text="General",
            icon=ft.icons.SETTINGS,
            content=ft.Column(
                controls=[
                    ft.Text(),
                    ft.TextField(
                        value=await self.storage.get("max_messages_per_chat"),
                        label="Max. messages per chat",
                        width=500,
                        on_change=self._max_messages_change,
                    ),
                    ft.Text(),
                    ft.TextField(
                        value=await self.storage.get("chat_update_rate"),
                        label="Chat UI Update rate(lower = higher CPU usage):",
                        width=500,
                        on_change=self._chat_update_rate_change,
                    ),
                ],
            ),
        )

    async def _get_appearance_tab(self) -> ft.Tab:
        return ft.Tab(
            text="Appearance",
            icon=ft.icons.BRUSH,
            content=ft.Column(
                controls=[
                    ft.Text(),
                    ft.Row(
                        controls=[
                            ft.Text(
                                "Chat font size",
                                size=16,
                            ),
                            ft.Slider(
                                value=await self.storage.get("chat_font_size"),
                                min=10,
                                max=50,
                                divisions=40,
                                label="{value}",
                                on_change_end=self._font_size_change,
                                expand=True,
                            ),
                        ],
                        alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
                    ),
                    ft.Text(),
                    ft.Row(
                        controls=[
                            ft.Text("Theme", size=16),
                            ft.Dropdown(
                                value=await self.storage.get("theme"),
                                options=[
                                    ft.dropdown.Option("System"),
                                    ft.dropdown.Option("Dark mode"),
                                    ft.dropdown.Option("Light mode"),
                                ],
                                width=200,
                                on_change=self._theme_select,
                            ),
                        ],
                        alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
                    ),
                    ft.Text(),
                    ft.Row(
                        controls=[
                            ft.Text("Chat color cycling", size=16),
                            ft.Checkbox(
                                value=await self.storage.get("color_switcher"),
                                label_position=ft.LabelPosition.LEFT,
                                on_change=self._on_color_switcher_click,
                            ),
                        ],
                        alignment=ft.MainAxisAlignment.SPACE_BETWEEN,
                    ),
                ],
            ),
        )

    async def _get_debug_tab(self) -> ft.Tab:
        return ft.Tab(
            text="Debug",
            icon=ft.icons.BUG_REPORT,
            content=ft.Column(
                controls=[
                    ft.Text(),
                    ft.Text("Log file location"),
                    ft.Row(
                        controls=[
                            ft.TextField(
                                value=str(LOG_PATH.absolute()),
                                read_only=True,
                                expand=True,
                            ),
                            ft.IconButton(
                                icon=ft.icons.COPY, on_click=self._log_path_copy_click
                            ),
                        ]
                    ),
                ]
            ),
        )

    async def _on_color_switcher_click(self, e):
        await self.storage.set("color_switcher", e.control.value)

    async def _theme_select(self, e):
        match e.data:
            case "System":
                self.page.theme_mode = ft.ThemeMode.SYSTEM
            case "Dark mode":
                self.page.theme_mode = ft.ThemeMode.DARK
            case "Light mode":
                self.page.theme_mode = ft.ThemeMode.LIGHT
            case _:
                pass

        await self.storage.set("theme", e.data)
        await self.page.update_async()

    async def _log_path_copy_click(self, _):
        await self.page.set_clipboard_async(str(LOG_PATH.absolute()))

    async def _max_messages_change(self, e):
        try:
            value = int(e.control.value)

            if value < 10 or value > 500:
                raise ValueError

            await self.storage.set("max_messages_per_chat", value)
            e.control.error_text = ""
            logging.debug(f"Updated max_messages_per_chat to {value}")

        except ValueError:
            e.control.error_text = "Value must be an integer between 10 and 500!"

        finally:
            await self.page.update_async()

    async def _back_click(self, _):
        self.page.views.pop()
        await self.page.update_async()

    async def _font_size_change(self, e):
        await self.storage.set("chat_font_size", e.control.value)
        await self.font_size_pubsub.send(e.control.value)
        await self.page.update_async()

    async def _chat_update_rate_change(self, e):
        try:
            value = float(e.control.value)

            if value < 0.3 or value > 1:
                raise ValueError

            e.control.error_text = ""
            await self.storage.set("chat_update_rate", e.control.value)
            logging.debug(f"Set chat_update_rate to {value}")

        except ValueError:
            e.control.error_text = "Value must be a decimal between 0.3 and 1."

        finally:
            await self.page.update_async()



import asyncio
import logging
from enum import Enum, auto
from math import isclose

import flet as ft

from hasherino.components.chat_message import ChatMessage
from hasherino.hasherino_dataclasses import Message
from hasherino.pubsub import PubSub
from hasherino.storage import AsyncKeyValueStorage


class ChatContainer(ft.Container):
    class _UiUpdateType(Enum):
        NO_UPDATE = (auto(),)
        SCROLL = (auto(),)
        PAGE = (auto(),)

    def __init__(
        self,
        persistent_storage: AsyncKeyValueStorage,
        memory_storage: AsyncKeyValueStorage,
        font_size_pubsub: PubSub,
    ):
        self.persistent_storage = persistent_storage
        self.memory_storage = memory_storage
        self.font_size_pubsub = font_size_pubsub
        self.is_chat_scrolled_down = False
        self.chat = ft.Column(
            expand=True,
            spacing=0,
            run_spacing=0,
            scroll=ft.ScrollMode.ALWAYS,
            auto_scroll=False,
            on_scroll=self.on_scroll,
        )
        super().__init__(
            content=self.chat,
            border=ft.border.all(1, ft.colors.OUTLINE),
            border_radius=5,
            padding=10,
            expand=True,
        )
        self.scheduled_ui_update: self._UiUpdateType = self._UiUpdateType.NO_UPDATE
        asyncio.ensure_future(self.update_ui())

    async def update_ui(self):
        while True:
            match self.scheduled_ui_update:
                case self._UiUpdateType.SCROLL:
                    await self.chat.scroll_to_async(offset=-1, duration=10)
                case self._UiUpdateType.PAGE:
                    await self.page.update_async()
                case self._UiUpdateType.NO_UPDATE | _:
                    pass

            self.scheduled_ui_update = self._UiUpdateType.NO_UPDATE

            await asyncio.sleep(
                float(await self.persistent_storage.get("chat_update_rate"))
            )

    async def on_scroll(self, event: ft.OnScrollEvent):
        self.is_chat_scrolled_down = isclose(
            event.pixels, event.max_scroll_extent, rel_tol=0.01
        )

    async def add_author_to_user_list(self, author: str):
        tab_name = await self.persistent_storage.get("channel")

        # Get existing list from memory or initialize a new one
        if user_list := await self.memory_storage.get("channel_user_list"):
            if tab_name in user_list:
                user_list[tab_name].append(author)
            else:
                user_list[tab_name] = [author]
        else:
            user_list = {tab_name: [author]}

        logging.debug(f"User {author} added to hash_table's user list")

        await self.memory_storage.set("channel_user_list", user_list)

    async def on_message(self, message: Message):
        if message.message_type == "chat_message":
            m = ChatMessage(
                message, self.page, await self.persistent_storage.get("chat_font_size")
            )
            await self.add_author_to_user_list(message.user.name)
            await m.subscribe_to_font_size_change(self.font_size_pubsub)

        elif message.message_type == "login_message":
            m = ft.Text(
                message.elements[0],
                italic=True,
                size=await self.persistent_storage.get("chat_font_size"),
            )

        self.chat.controls.append(m)

        n_messages_to_remove = len(
            self.chat.controls
        ) - await self.persistent_storage.get("max_messages_per_chat")
        if n_messages_to_remove > 0:
            del self.chat.controls[:n_messages_to_remove]
            logging.debug(
                f"Chat has {len(self.chat.controls)} lines in it, removed {n_messages_to_remove}"
            )

        if self.is_chat_scrolled_down:
            self.scheduled_ui_update = self._UiUpdateType.SCROLL
        elif (
            self.scheduled_ui_update != self._UiUpdateType.SCROLL
        ):  # Scroll already updates
            self.scheduled_ui_update = self._UiUpdateType.PAGE



from dataclasses import dataclass
from enum import Enum
from json import dumps


@dataclass
class Badge:
    id: str
    name: str
    url: str


@dataclass
class HasherinoUser:
    name: str
    badges: list[Badge] | None = None
    chat_color: str | None = None


class EmoteSource(Enum):
    TWITCH = 0
    SEVENTV = 1


@dataclass
class Emote:
    name: str
    id: str
    url: str


@dataclass
class Message:
    user: HasherinoUser
    elements: list[str | Emote]
    message_type: str
    me: bool


# print(User("asd").toJSON())
# pass



import asyncio
import logging

import flet as ft

from hasherino.api import helix
from hasherino.api.seven_tv import SevenTV
from hasherino.storage import AsyncKeyValueStorage
from hasherino.twitch_websocket import TwitchWebsocket


class HasherinoTab(ft.Tab):
    def __init__(
        self,
        channel: str,
        persistent_storage: AsyncKeyValueStorage,
        memory_storage: AsyncKeyValueStorage,
    ):
        super().__init__(tab_content=ft.Row(controls=[ft.Text(channel)]))
        self.persistent_storage = persistent_storage
        self.memory_storage = memory_storage
        self.channel = channel

    async def _get_channel_seventv_emotes(self, user: helix.TwitchUser) -> dict:
        try:
            seventv_user = await SevenTV.get_user(user.id)

            if not seventv_user:
                return dict()

            ttv_connection = next(
                connection
                for connection in seventv_user["connections"]
                if connection["platform"] == "TWITCH"
            )
            active_ttv_set = next(
                ttv_set
                for ttv_set in seventv_user["emote_sets"]
                if ttv_set["id"] == ttv_connection["emote_set_id"]
            )
            logging.info(
                f"Loaded {len(active_ttv_set['emotes'])} 7tv emotes for {self.channel}"
            )
            return {
                emote["name"]: emote["data"]["id"] for emote in active_ttv_set["emotes"]
            }
        except Exception as e:
            raise Exception(
                f"Failed to load 7tv emotes for {self.channel} with error {e}"
            ) from e

    async def _get_channel_ttv_emotes(
        self, app_id: str, helix_token: str, user: helix.TwitchUser
    ):
        try:
            return await helix.get_channel_emotes(app_id, helix_token, str(user.id))
        except Exception as e:
            raise Exception(
                f"Failed to load ttv emotes for {self.channel} with error {e}"
            ) from e

    async def load_emotes(self):
        try:
            async with asyncio.TaskGroup() as tg:
                app_id_task = tg.create_task(self.persistent_storage.get("app_id"))
                helix_token_task = tg.create_task(self.persistent_storage.get("token"))

            app_id = await app_id_task
            token = await helix_token_task

            user: helix.TwitchUser = (
                await helix.get_users(
                    app_id,
                    token,
                    [self.channel],
                )
            )[0]

            seventv_emotes = await self._get_channel_seventv_emotes(user)
            logging.debug(f"Loaded 7tv emotes for {user.login}: {seventv_emotes}")

            if not (emotes := await self.memory_storage.get("7tv_emotes")):
                emotes = dict()

            emotes[user.login] = seventv_emotes
            await self.memory_storage.set("7tv_emotes", emotes)

        except Exception as e:
            logging.error(f"Error while loading emotes: {e}")


class Tabs(ft.Tabs):
    def __init__(
        self,
        memory_storage: AsyncKeyValueStorage,
        persistent_storage: AsyncKeyValueStorage,
    ):
        super().__init__(
            tabs=[],
            on_change=self.change,
        )
        self.memory_storage = memory_storage
        self.persistent_storage = persistent_storage

    async def add_tab(self, channel: str):
        tab = HasherinoTab(channel, self.persistent_storage, self.memory_storage)
        asyncio.ensure_future(tab.load_emotes())
        close_button = ft.IconButton(icon=ft.icons.CLOSE, on_click=self.close)
        close_button.parent_tab = tab
        tab.tab_content.controls.append(close_button)
        self.tabs = [tab]
        logging.info(f"Added tab {channel}")
        await self.page.add_async()

    async def close(self, button_click: ft.ControlEvent):
        websocket: TwitchWebsocket = await self.memory_storage.get("websocket")
        tab_channel = button_click.control.parent_tab.channel
        await websocket.leave_channel(tab_channel)
        self.tabs.remove(button_click.control.parent_tab)
        await self.persistent_storage.set("channel", None)
        logging.info(f"Closed tab {tab_channel}")
        await self.page.add_async()

    async def change(self, e):
        tab = e.control.tabs[e.control.selected_index]



import ssl

import certifi
from aiohttp import ClientSession, TCPConnector


class SevenTV:
    @staticmethod
    def _get_client_session() -> ClientSession:
        ssl_context = ssl.create_default_context(cafile=certifi.where())
        conn = TCPConnector(ssl=ssl_context)
        return ClientSession(connector=conn)

    @staticmethod
    async def _gql_request(query: dict):
        async with SevenTV._get_client_session() as session:
            async with session.post(
                f"https://7tv.io/v3/gql",
                headers={
                    "Content-Type": "application/json",
                },
                json=query,
            ) as response:
                return (await response.json())["data"]

    @staticmethod
    async def get_user(ttv_user_id: str):
        return (
            await SevenTV._gql_request(
                {
                    "operationName": "GetUserByConnection",
                    "query": """
            query GetUserByConnection($platform: ConnectionPlatform! $id: String!) {
                userByConnection (platform: $platform id: $id) {
                    id
                    type
                    username
                    roles
                    created_at
                    connections {
                        id
                        platform
                        emote_set_id
                    }
                    editors {
                      user {
                        id
                        username
                      }
                    }
                    emote_sets {
                        id
                        emotes {
                            id
                            name
                            data {
                              id
                              name
                            }
                        }
                        capacity
                    }
                }
            }""",
                    "variables": {
                        "platform": "TWITCH",
                        "id": ttv_user_id,
                    },
                }
            )
        )["userByConnection"]
